﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.InteropServices;
using System.Text;

namespace Inedo.WindowsServices
{
    /// <summary>
    /// Represents a Windows service on the local machine.
    /// </summary>
    [Undisclosed]
    public sealed class WindowsService : IDisposable
    {
        private IntPtr serviceHandle;
        private IntPtr managerHandle;
        private bool disposed;
        private Lazy<string> displayName;
        private Lazy<string> description;
        private Lazy<ServiceFailureActions> failureActions;
        private Lazy<ReadOnlyCollection<string>> dependencies;

        private WindowsService()
        {
            this.displayName = new Lazy<string>(this.GetServiceDisplayName);
            this.description = new Lazy<string>(this.GetServiceDescription);
            this.failureActions = new Lazy<ServiceFailureActions>(this.GetFailureActions);
            this.dependencies = new Lazy<ReadOnlyCollection<string>>(this.GetDependencies);
        }

        /// <summary>
        /// Finalizes an instance of the <see cref="WindowsService"/> class.
        /// </summary>
        ~WindowsService()
        {
            this.Dispose(false);
        }

        /// <summary>
        /// Gets the name of the service.
        /// </summary>
        public string Name { get; private set; }
        /// <summary>
        /// Gets the display name of the service.
        /// </summary>
        public string DisplayName
        {
            get { return this.displayName.Value; }
            set { this.SetDisplayName(value); }
        }
        /// <summary>
        /// Gets or sets the service's description.
        /// </summary>
        public string Description
        {
            get { return this.description.Value; }
            set { this.SetServiceDescription(value); }
        }
        /// <summary>
        /// Gets or sets the failure actions to perform if the service crashes.
        /// </summary>
        public ServiceFailureActions FailureActions
        {
            get { return this.failureActions.Value; }
            set { this.SetFailureActions(value); }
        }
        /// <summary>
        /// Gets or sets a value indicating whether the delayed auto start feature is enabled for this service.
        /// </summary>
        public bool DelayedStart
        {
            get { return this.GetDelayedAutoStart(); }
            set { this.SetDelayedAutoStart(value); }
        }
        /// <summary>
        /// Gets or sets the file name of the service process.
        /// </summary>
        public string FileName
        {
            get { return this.GetBinaryPathName(); }
            set { this.SetBinaryPathName(value); }
        }
        /// <summary>
        /// Gets the user account name of the service.
        /// </summary>
        public string UserAccountName => this.GetUserAccountName();
        /// <summary>
        /// Gets or sets the startup mode of the service.
        /// </summary>
        public ServiceStartMode StartMode
        {
            get { return this.GetStartMode(); }
            set { this.SetStartMode(value); }
        }

        /// <summary>
        /// Gets the service dependencies.
        /// </summary>
        public ReadOnlyCollection<string> Dependencies => this.dependencies.Value;

        /// <summary>
        /// Returns the service on the local machine with the specified name.
        /// </summary>
        /// <param name="serviceName">The name of the service.</param>
        /// <returns>The service with the specified name or null if the service was not found.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="serviceName"/> is null or empty.</exception>
        /// <exception cref="Win32Exception">The service controller could not be opened.</exception>
        public static WindowsService GetService(string serviceName)
        {
            if (string.IsNullOrEmpty(serviceName))
                throw new ArgumentNullException(nameof(serviceName));

            var scm = ServiceApi.OpenSCManager(null, null, ServiceApi.SCM_ACCESS.SC_MANAGER_ALL_ACCESS);
            if (scm == IntPtr.Zero)
                throw new Win32Exception();

            try
            {
                var svc = ServiceApi.OpenService(scm, serviceName, ServiceApi.SERVICE_ACCESS.SERVICE_ALL_ACCESS);
                if (svc == IntPtr.Zero)
                    return null;

                return new WindowsService
                {
                    Name = serviceName,
                    serviceHandle = svc,
                    managerHandle = scm
                };
            }
            catch
            {
                ServiceApi.CloseServiceHandle(scm);
                throw;
            }
        }

        /// <summary>
        /// Creates a service and adds it to the local service controller database.
        /// </summary>
        /// <param name="serviceName">The name of the service.</param>
        /// <param name="displayName">The display name of the service.</param>
        /// <param name="fileName">The full path of the service's executable file.</param>
        /// <param name="userAccount">The user account the service will run as. See Remarks.</param>
        /// <param name="password">The password for the <paramref name="userAccount"/> parameter.</param>
        /// <returns>The newly-created service.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="serviceName"/>, <paramref name="displayName"/>, or <paramref name="fileName"/> is null or empty.</exception>
        /// <exception cref="Win32Exception">The service controller could not be opened or the service could not be created.</exception>
        /// <remarks>
        /// <para><paramref name="userAccount"/> is specified in the format: <c>domain\user</c></para>
        /// <para>Built-in accounts should be prefixed with <c>NT AUTHORITY</c>. For example: <c>NT AUTHORITY\LocalSystem</c>.</para>
        /// </remarks>
        public static WindowsService CreateService(string serviceName, string displayName, string fileName, string userAccount, string password)
        {
            if (string.IsNullOrEmpty(serviceName))
                throw new ArgumentNullException(nameof(serviceName));
            if (string.IsNullOrEmpty(displayName))
                throw new ArgumentNullException(nameof(displayName));
            if (string.IsNullOrEmpty(fileName))
                throw new ArgumentNullException(nameof(fileName));

            AH.UserAccounts.CanonicalizeAccountName(ref userAccount);

            if (string.IsNullOrEmpty(userAccount) || userAccount == "NT AUTHORITY\\LocalSystem")
                userAccount = null;

            var scm = ServiceApi.OpenSCManager(null, null, ServiceApi.SCM_ACCESS.SC_MANAGER_ALL_ACCESS);
            if (scm == IntPtr.Zero)
                throw new Win32Exception();

            if (!string.IsNullOrEmpty(userAccount) && !userAccount.StartsWith("NT AUTHORITY"))
            {
                try { AddServiceLogonRight(userAccount); }
                catch { }
            }

            try
            {
                var svc = ServiceApi.CreateService(
                    scm,
                    serviceName,
                    displayName,
                    ServiceApi.SERVICE_ACCESS.SERVICE_ALL_ACCESS,
                    ServiceApi.SERVICE_TYPE.SERVICE_WIN32_OWN_PROCESS,
                    ServiceApi.SERVICE_START.SERVICE_AUTO_START,
                    ServiceApi.SERVICE_ERROR.SERVICE_ERROR_NORMAL,
                    fileName,
                    null,
                    null,
                    null,
                    userAccount,
                    password
                );

                if (svc == IntPtr.Zero)
                    throw new Win32Exception();

                return new WindowsService
                {
                    Name = serviceName,
                    serviceHandle = svc,
                    managerHandle = scm
                };
            }
            catch
            {
                ServiceApi.CloseServiceHandle(scm);
                throw;
            }
        }

        /// <summary>
        /// Creates a service running as LocalSystem and adds it to the local service controller database.
        /// </summary>
        /// <param name="serviceName">The name of the service.</param>
        /// <param name="displayName">The display name of the service.</param>
        /// <param name="fileName">The full path of the service's executable file.</param>
        /// <returns>The newly-created service.</returns>
        public static WindowsService CreateService(string serviceName, string displayName, string fileName)
        {
            return CreateService(serviceName, displayName, fileName, null, null);
        }

        /// <summary>
        /// Deletes the service from the service controller.
        /// </summary>
        public void Delete()
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            if (!ServiceApi.DeleteService(this.serviceHandle))
                throw new Win32Exception();
        }

        /// <summary>
        /// Changes the service's user account name and password.
        /// </summary>
        /// <param name="userAccountName">The user account to change the service to.</param>
        /// <param name="password">The password of the user account.</param>
        public void SetUserAccount(string userAccountName, string password)
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            if (!ServiceApi.ChangeServiceConfig(this.serviceHandle, lpServiceStartName: userAccountName, lpPassword: password))
                throw new Win32Exception();
        }

        /// <summary>
        /// Sets the service's dependencies.
        /// </summary>
        /// <param name="dependencies">The dependencies.</param>
        public void SetDependencies(IList<string> dependencies)
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));
            if (dependencies == null)
                throw new ArgumentNullException(nameof(dependencies));

            string value = string.Join("\0", dependencies) + "\0";

            if (!ServiceApi.ChangeServiceConfig(this.serviceHandle, lpDependencies: value))
                throw new Win32Exception();

            this.dependencies = new Lazy<ReadOnlyCollection<string>>(this.GetDependencies);
        }

        /// <summary>
        /// Sets the display name.
        /// </summary>
        /// <param name="displayName">The display name.</param>
        private void SetDisplayName(string displayName)
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            if (!ServiceApi.ChangeServiceConfig(this.serviceHandle, lpDisplayName: displayName))
                throw new Win32Exception();

            this.displayName = new Lazy<string>(this.GetServiceDisplayName);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        private string GetServiceDisplayName()
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            var buffer = new StringBuilder(2048);
            uint size = 2048;
            ServiceApi.GetServiceDisplayName(this.serviceHandle, this.Name, buffer, ref size);
            return buffer.ToString();
        }

        private string GetServiceDescription()
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            var buffer = IntPtr.Zero;
            try
            {
                buffer = this.QueryServiceConfig(ServiceApi.SERVICE_CONFIG.DESCRIPTION);
                if (buffer == IntPtr.Zero)
                    return null;

                unsafe
                {
                    var s = Marshal.PtrToStringAnsi(*(IntPtr*)buffer.ToPointer());
                    if (s == string.Empty)
                        return null;

                    return s;
                }
            }
            finally
            {
                if (buffer != IntPtr.Zero)
                    Marshal.FreeCoTaskMem(buffer);
            }
        }
        private void SetServiceDescription(string value)
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            unsafe
            {
                IntPtr buffer = IntPtr.Zero;
                try
                {
                    buffer = Marshal.StringToCoTaskMemAnsi(value ?? string.Empty);
                    if (!ServiceApi.ChangeServiceConfig2(this.serviceHandle, ServiceApi.SERVICE_CONFIG.DESCRIPTION, new IntPtr(&buffer)))
                        throw new Win32Exception();

                    this.description = new Lazy<string>(this.GetServiceDescription);
                }
                finally
                {
                    if (buffer != IntPtr.Zero)
                        Marshal.FreeCoTaskMem(buffer);
                }
            }
        }

        private ServiceFailureActions GetFailureActions()
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            var buffer = IntPtr.Zero;
            try
            {
                buffer = this.QueryServiceConfig(ServiceApi.SERVICE_CONFIG.FAILURE_ACTIONS);
                if (buffer == IntPtr.Zero)
                    return null;

                unsafe
                {
                    var ptr = (ServiceApi.SERVICE_FAILURE_ACTIONS*)buffer.ToPointer();
                    var actions = new List<ServiceControllerAction>((int)ptr->cActions);
                    for (uint i = 0; i < ptr->cActions; i++)
                    {
                        actions.Add(
                            new ServiceControllerAction(
                                (ServiceControllerActionType)ptr->lpsaActions[i].Type,
                                new TimeSpan(0, 0, 0, 0, (int)ptr->lpsaActions[i].Delay)
                            )
                        );
                    }

                    return new ServiceFailureActions(
                        ptr->dwResetPeriod != 0 ? (int?)ptr->dwResetPeriod : null,
                        ptr->lpRebootMsg != IntPtr.Zero ? Marshal.PtrToStringAnsi(ptr->lpRebootMsg) : null,
                        ptr->lpCommand != IntPtr.Zero ? Marshal.PtrToStringAnsi(ptr->lpCommand) : null,
                        actions.ToArray()
                    );
                }
            }
            finally
            {
                if (buffer != IntPtr.Zero)
                    Marshal.FreeCoTaskMem(buffer);
            }
        }
        private void SetFailureActions(ServiceFailureActions value)
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            var rebootPtr = IntPtr.Zero;
            var commandPtr = IntPtr.Zero;
            try
            {
                rebootPtr = Marshal.StringToCoTaskMemAnsi(value.RebootMessage ?? string.Empty);
                commandPtr = Marshal.StringToCoTaskMemAnsi(value.Command ?? string.Empty);

                unsafe
                {
                    var actionData = stackalloc ServiceApi.SC_ACTION[value.Actions.Count];
                    for (int i = 0; i < value.Actions.Count; i++)
                    {
                        actionData[i] = new ServiceApi.SC_ACTION
                        {
                            Type = (ServiceApi.SC_ACTION_TYPE)value.Actions[i].Type,
                            Delay = (uint)value.Actions[i].Delay.TotalMilliseconds
                        };
                    }

                    var data = new ServiceApi.SERVICE_FAILURE_ACTIONS
                    {
                        dwResetPeriod = (uint)(value.ResetPeriod ?? 0),
                        lpRebootMsg = rebootPtr,
                        lpCommand = commandPtr,
                        cActions = (uint)value.Actions.Count,
                        lpsaActions = value.Actions.Count > 0 ? actionData : null
                    };

                    if (!ServiceApi.ChangeServiceConfig2(this.serviceHandle, ServiceApi.SERVICE_CONFIG.FAILURE_ACTIONS, new IntPtr(&data)))
                        throw new Win32Exception();

                    this.failureActions = new Lazy<ServiceFailureActions>(this.GetFailureActions);
                }
            }
            finally
            {
                if (commandPtr != IntPtr.Zero)
                    Marshal.FreeCoTaskMem(commandPtr);
                if (rebootPtr != IntPtr.Zero)
                    Marshal.FreeCoTaskMem(rebootPtr);
            }
        }

        private bool GetDelayedAutoStart()
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            unsafe
            {
                uint value = 0;
                uint size;
                ServiceApi.QueryServiceConfig2(this.serviceHandle, ServiceApi.SERVICE_CONFIG.DELAYED_AUTO_START_INFO, new IntPtr(&value), sizeof(uint), out size);
                return value != 0;
            }
        }
        private void SetDelayedAutoStart(bool value)
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            unsafe
            {
                uint intValue = value ? 1u : 0u;
                ServiceApi.ChangeServiceConfig2(this.serviceHandle, ServiceApi.SERVICE_CONFIG.DELAYED_AUTO_START_INFO, new IntPtr(&value));
            }
        }

        private ReadOnlyCollection<string> GetDependencies()
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            unsafe
            {
                fixed (byte* buffer = new byte[8192])
                {
                    var info = (ServiceApi.QUERY_SERVICE_CONFIG*)buffer;
                    uint bytesNeeded;
                    if (ServiceApi.QueryServiceConfig(this.serviceHandle, info, 8192, out bytesNeeded))
                    {
                        if (info->lpDependencies == IntPtr.Zero)
                            return null;

                        var rubbish = new List<string>();
                        var chars = (char*)info->lpDependencies.ToPointer();
                        var s = new string(chars);
                        while (s.Length > 0)
                        {
                            rubbish.Add(s);
                            chars += s.Length + 1;
                            s = new string(chars);
                        }

                        return rubbish.AsReadOnly();
                    }
                    else
                    {
                        return new List<string>().AsReadOnly();
                    }
                }
            }
        }

        private string GetBinaryPathName()
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            unsafe
            {
                fixed (byte* buffer = new byte[8192])
                {
                    var info = (ServiceApi.QUERY_SERVICE_CONFIG*)buffer;
                    uint bytesNeeded;
                    if (ServiceApi.QueryServiceConfig(this.serviceHandle, info, 8192, out bytesNeeded))
                    {
                        if (info->lpBinaryPathName == IntPtr.Zero)
                            return null;

                        return Marshal.PtrToStringUni(info->lpBinaryPathName);
                    }
                    else
                    {
                        return null;
                    }
                }
            }
        }
        private void SetBinaryPathName(string path)
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            ServiceApi.ChangeServiceConfig(
                this.serviceHandle,
                lpBinaryPathName: path
            );
        }

        private string GetUserAccountName()
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            unsafe
            {
                fixed (byte* buffer = new byte[8192])
                {
                    var info = (ServiceApi.QUERY_SERVICE_CONFIG*)buffer;
                    uint bytesNeeded;
                    if (ServiceApi.QueryServiceConfig(this.serviceHandle, info, 8192, out bytesNeeded))
                    {
                        if (info->lpServiceStartName == IntPtr.Zero)
                            return null;

                        return Marshal.PtrToStringUni(info->lpServiceStartName);
                    }
                    else
                    {
                        return null;
                    }
                }
            }
        }

        private ServiceStartMode GetStartMode()
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            unsafe
            {
                fixed (byte* buffer = new byte[8192])
                {
                    var info = (ServiceApi.QUERY_SERVICE_CONFIG*)buffer;
                    uint bytesNeeded;
                    ServiceApi.QueryServiceConfig(this.serviceHandle, info, 8192, out bytesNeeded);
                    return (ServiceStartMode)info->dwStartType;
                }
            }
        }
        private void SetStartMode(ServiceStartMode mode)
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            if (!ServiceApi.ChangeServiceConfig(this.serviceHandle, dwStartType: (uint)mode))
                throw new Win32Exception();
        }

        private IntPtr QueryServiceConfig(ServiceApi.SERVICE_CONFIG config)
        {
            if (this.disposed)
                throw new ObjectDisposedException(nameof(WindowsService));

            uint bufferSize;
            ServiceApi.QueryServiceConfig2(this.serviceHandle, config, IntPtr.Zero, 0, out bufferSize);
            if (bufferSize > 0)
            {
                var buffer = IntPtr.Zero;
                try
                {
                    buffer = Marshal.AllocCoTaskMem((int)bufferSize);
                    if (!ServiceApi.QueryServiceConfig2(this.serviceHandle, config, buffer, bufferSize, out bufferSize))
                        throw new Win32Exception();

                    return buffer;
                }
                catch
                {
                    if (buffer != IntPtr.Zero)
                        Marshal.FreeCoTaskMem(buffer);

                    throw;
                }
            }

            return IntPtr.Zero;
        }

        private void Dispose(bool disposing)
        {
            if (!this.disposed)
            {
                ServiceApi.CloseServiceHandle(this.serviceHandle);
                ServiceApi.CloseServiceHandle(this.managerHandle);
                this.disposed = true;
            }
        }

        private static void AddServiceLogonRight(string userAccount)
        {
            var sid = new byte[1000];
            uint sidLength = 1000;
            var domainName = new StringBuilder(200);
            uint domainNameLength = 200;
            uint use;
            if (!NativeMethods.LookupAccountName(null, userAccount, sid, ref sidLength, domainName, ref domainNameLength, out use))
                throw new ArgumentException("Invalid or unknown user account.");

            unsafe
            {
                var attr = new NativeMethods.LSA_OBJECT_ATTRIBUTES();
                var policyHandle = IntPtr.Zero;

                try
                {
                    if (NativeMethods.LsaOpenPolicy(null, &attr, 28, out policyHandle) != 0)
                        throw new InvalidOperationException("Could not open LSA policy.");

                    var privilegeName = new NativeMethods.LSA_UNICODE_STRING
                    {
                        Length = (ushort)("SeServiceLogonRight".Length * sizeof(char)),
                        MaximumLength = (ushort)(("SeServiceLogonRight".Length + 1) * sizeof(char)),
                        Buffer = Marshal.StringToCoTaskMemUni("SeServiceLogonRight")
                    };

                    try
                    {
                        var res = NativeMethods.LsaAddAccountRights(policyHandle, sid, &privilegeName, 1);
                        if (res != 0)
                            throw new InvalidOperationException("Could not add SeServiceLogonRight.");
                    }
                    finally
                    {
                        if (privilegeName.Buffer != IntPtr.Zero)
                            Marshal.FreeCoTaskMem(privilegeName.Buffer);
                    }
                }
                finally
                {
                    if (policyHandle != IntPtr.Zero)
                        NativeMethods.LsaClose(policyHandle);
                }
            }
        }

        private static class NativeMethods
        {
            [DllImport("advapi32.dll", SetLastError = true, CharSet = CharSet.Unicode)]
            [return: MarshalAs(UnmanagedType.Bool)]
            public static extern bool LookupAccountName(
                [MarshalAs(UnmanagedType.LPWStr)] string lpSystemName,
                [MarshalAs(UnmanagedType.LPWStr)] string lpAccountName,
                [MarshalAs(UnmanagedType.LPArray)] byte[] Sid,
                ref uint sbCid,
                [MarshalAs(UnmanagedType.LPWStr)] StringBuilder ReferencedDomainName,
                ref uint cchReferencedDomainName,
                out uint peUse);

            [DllImport("advapi32.dll", SetLastError = false)]
            public static unsafe extern uint LsaOpenPolicy(
               LSA_UNICODE_STRING* SystemName,
               LSA_OBJECT_ATTRIBUTES* ObjectAttributes,
               uint DesiredAccess,
               out IntPtr PolicyHandle);

            [DllImport("advapi32.dll", SetLastError = false)]
            public static unsafe extern uint LsaAddAccountRights(
                IntPtr PolicyHandle,
                [MarshalAs(UnmanagedType.LPArray)] byte[] AccountSid,
                LSA_UNICODE_STRING* UserRights,
                uint CountOfRights);

            [DllImport("advapi32.dll", SetLastError = false)]
            public static extern uint LsaClose(IntPtr ObjectHandle);

            [StructLayout(LayoutKind.Sequential)]
            public struct LSA_UNICODE_STRING
            {
                public ushort Length;
                public ushort MaximumLength;
                public IntPtr Buffer;
            }

            [StructLayout(LayoutKind.Sequential)]
            public struct LSA_OBJECT_ATTRIBUTES
            {
                public uint Length;
                public IntPtr RootDirectory;
                public IntPtr ObjectName;
                public uint Attributes;
                public IntPtr SecurityDescriptor;
                public IntPtr SecurityQualityOfService;
            }
        }
    }
}
